#include "MaskBuffer.h"
#include "XRSurfaceData.h"
#include "X11SurfaceData.h"
#include "Trace.h"
#include "awt_GraphicsEnv.h"

#define MAX(a,b) ((a) > (b) ? (a) : (b))
#define MIN(a,b) ((a) < (b) ? (a) : (b))

#define GET_DIRTY_LINE_REGION(X1, Y1, X2, Y2, RX1, RY1, RX2, RY2)  {     \
    if(X1 < X2) {                                                        \
      RX1 = X1;                                                          \
      RX2 = X2;                                                          \
    }else {                                                              \
      RX1 = X2;                                                          \
      RX2 = X1;                                                          \
    }                                                                    \
                                                                         \
    if(Y1 < Y2) {                                                        \
      RY1 = Y1;                                                          \
      RY2 = Y2;                                                          \
    }else {                                                              \
      RY1 = Y2;                                                          \
      RY2 = Y1;                                                          \
    }                                                                    \
  }

#ifndef HEADLESS

MaskBuffer* initMaskBuffer(Window window) {
  MaskBuffer *buffer = malloc(sizeof(MaskBuffer));
  initRegion(&buffer->region);
 
  initXRList(&buffer->rects, sizeof(XRectangle), 512);
  initXRList(&buffer->lines, sizeof(XSegment), 512);

  buffer->xTiles = 0;
  buffer->yTiles = 0;
  buffer->xTileCap = 0;
  buffer->yTileCap = 0;
  buffer->tiles = NULL;

  XRenderPictFormat *fmt = XRenderFindStandardFormat(awt_display, PictStandardA8);

  buffer->maskPixmap = XCreatePixmap(awt_display, window, MASK_TILE_SIZE, MASK_TILE_SIZE, 8);
  buffer->lineMaskPixmap = XCreatePixmap(awt_display, window, MASK_TILE_SIZE, MASK_TILE_SIZE, 8);

  XRenderPictureAttributes pict_attr;
  buffer->maskPicture = XRenderCreatePicture(awt_display, buffer->maskPixmap, fmt, 0, &pict_attr);
  buffer->lineMaskPicture = XRenderCreatePicture(awt_display, buffer->lineMaskPixmap, fmt, 0, &pict_attr);

  XRenderFillRectangle (awt_display, PictOpClear, buffer->maskPicture, &color_black, 0, 0, MASK_TILE_SIZE, MASK_TILE_SIZE);
  XRenderFillRectangle (awt_display, PictOpClear, buffer->lineMaskPicture, &color_black, 0, 0, MASK_TILE_SIZE, MASK_TILE_SIZE);

/*
  //1x1 image mask pixmap
  Pixmap m1p =  XCreatePixmap(awt_display, window, 1, 1, 8);
  buffer->img1Mask = XRenderCreatePicture(awt_display, m1p, fmt, 0, &pict_attr);
  XRenderFillRectangle (awt_display, PictOpSrc, buffer->img1Mask, &color_black, 0, 0, 1, 1);

  //8x8 image mask pixmap
  Pixmap m8p =  XCreatePixmap(awt_display, window, 8, 8, 8);
  buffer->img8Mask = XRenderCreatePicture(awt_display, m8p, fmt, 0, &pict_attr);
  XRenderFillRectangle (awt_display, PictOpSrc, buffer->img8Mask, &color_black, 0, 0, 8, 8);

  //64x64 image mask pixmap
  Pixmap m64p =  XCreatePixmap(awt_display, window, 64, 64, 8);
  buffer->img64Mask = XRenderCreatePicture(awt_display, m64p, fmt, 0, &pict_attr);
  XRenderFillRectangle (awt_display, PictOpSrc, buffer->img64Mask, &color_black, 0, 0, 64, 64);
*/

  Pixmap mp =  XCreatePixmap(awt_display, window, 64, 64, 8);
  buffer->maxMask = XRenderCreatePicture(awt_display, mp, fmt, 0, &pict_attr);
  XRenderFillRectangle (awt_display, PictOpClear, buffer->maxMask, &color_black, 0, 0, 64, 64);
  buffer->maxWidth = 64;
  buffer->maxHeight = 64;

  buffer->validatedGCAlpha = 1.0f;
  XGCValues values;
  values.graphics_exposures = 0;

  values.foreground = 255;
  buffer->drawLineGC = XCreateGC(awt_display, buffer->lineMaskPixmap, GCForeground | GCGraphicsExposures, &values);
  /*Invisible GC for readback assistance*/
  values.foreground = 0;
  buffer->clearLineGC = XCreateGC(awt_display, buffer->lineMaskPixmap, GCForeground | GCGraphicsExposures, &values);
  buffer->maskGC = XCreateGC(awt_display, buffer->maskPixmap, GCGraphicsExposures, &values);
 
  buffer->alphaData = malloc(32*32);
  buffer->alphaImg = XCreateImage(awt_display, &buffer->maskPixmap, 8, ZPixmap, 0, (char *) buffer->alphaData, 32, 32, 8, 0);
  buffer->aImgWidth = 32;
  buffer->aImgHeight = 32;

  return buffer;
}

static void allocTileBuffer(MaskBuffer* buf) {
  buf->xTiles = ((buf->region.x2 - buf->region.x) / MASK_TILE_SIZE) + 1;
  buf->yTiles = ((buf->region.y2 - buf->region.y) / MASK_TILE_SIZE) + 1;

  /* If current tile-array is too small, allocate a larger one */
  if(buf->tiles == NULL || buf->xTileCap < buf->xTiles || buf->yTileCap < buf->yTiles) {
    buf->xTileCap = MAX((buf->xTiles * 2), buf->xTileCap);
    buf->yTileCap = MAX((buf->yTiles * 2), buf->yTileCap);
    if(buf->tiles != NULL) {
      free(buf->tiles);
    }
    buf->tiles = calloc(buf->xTileCap * buf->yTileCap, sizeof(XrTile));
  }

  /* Init/Reset the tiles that will be used later */
  int x,y;
  for(y=0; y < buf->yTiles; y++) {
    for(x=0; x < buf->xTiles; x++) {
      initXrTile(&buf->tiles[y*buf->xTileCap + x]);
    }
 }
}

static void cleanMaskBuffer(MaskBuffer *buf) {
  initRegion(&buf->region);
  buf->xTiles = 0;
  buf->yTiles = 0;
  clearXRList(&buf->rects);
  clearXRList(&buf->lines);
}

static void getTileDirtyArea(MaskRegion *dirtyRectArea, MaskRegion *dirtyLineArea, MaskRegion *totalDirtyArea) {
  totalDirtyArea->x = MIN(dirtyRectArea->x, dirtyLineArea->x);
  totalDirtyArea->y = MIN(dirtyRectArea->y, dirtyLineArea->y);
  totalDirtyArea->x2 = MAX(dirtyRectArea->x2, dirtyLineArea->x2);
  totalDirtyArea->y2 = MAX(dirtyRectArea->y2, dirtyLineArea->y2);
}

static void growDirtyRegion(MaskRegion *region, jint x, jint y, jint x2, jint y2) {
  if(x < region->x){
    region->x = x;
  }
  if(y < region->y) {
    region->y = y;
  }
  if(x2 > region->x2){ 
    region->x2 = x2;
  }
  if(y2 > region->y2){
    region->y2 = y2;
  }
}

static void growDirtyRegionTileLimit(MaskRegion *region, jint x, jint y, jint x2, jint y2) {
  if(x < region->x){
    region->x = MAX(x, 0);
  }
  if(y < region->y) {
    region->y = MAX(y, 0);
  }
  if(x2 > region->x2){ 
    region->x2 = MIN(x2, MASK_TILE_SIZE);
  }
  if(y2 > region->y2){
    region->y2 = MIN(y2, MASK_TILE_SIZE);
  }
}

/**
* Limits the rect's coordinates to the mask coordinates.
* The result is used by growDirtyRegion.
*/
static void limitRetcCoords(XRectangle *rect)  {
   if((rect->x + rect->width) > MASK_TILE_SIZE) {
      rect->width = MASK_TILE_SIZE - rect->x;
   }
   if((rect->y + rect->height) > MASK_TILE_SIZE) {
      rect->height = MASK_TILE_SIZE - rect->y;
   }

  if(rect->x < 0){
    rect->width += rect->x;
    rect->x = 0;
  }
  if(rect->y < 0) {
     rect->height += rect->y;
     rect->y = 0;
  }
}

static void initRegion(MaskRegion *region) {
  region->x = 99999; /*TODO: Integer max/min*/
  region->y = 99999;
  region->x2 = -99999;
  region->y2 = -99999;
}

void addRect(MaskBuffer* buf, short x, short y, unsigned short w, unsigned short h) {
  XRectangle *rect = (XRectangle*) getNewXR(&buf->rects, sizeof(XRectangle));
  rect->x = x;
  rect->y = y;
  rect->width = w;
  rect->height = h;

  growDirtyRegion(&buf->region, x, y, x+w, y+h);
}

void addLine(MaskBuffer* buf, int x1, int y1, int x2, int y2) {
  /* EXA is not able to accalerate diagonal lines, 
     we try to "guide" it a bit to avoid excessive migration
     See project documentation for an detailed explanation */

  int minX, minY, maxX, maxY;
  GET_DIRTY_LINE_REGION(x1, y1, x2, y2, minX, minY, maxX, maxY);
  int xDiff = maxX - minX;
  int yDiff = maxY - minY;

  if(xDiff == 0 || yDiff == 0) {
     addRect(buf, minX, minY, maxX - minX + 1, maxY - minY + 1);
  } else if(xDiff == 1 && yDiff == 1) {
     addRect(buf, x1, y1, 1, 1);
     addRect(buf, x2, y2, 1, 1);
  } else {
     XSegment *line = (XSegment*) getNewXR(&buf->lines, sizeof(XSegment));
     line->x1 = x1;
     line->y1 = y1;
     line->x2 = x2;
     line->y2 = y2;

     growDirtyRegion(&buf->region, minX, minY, maxX + 1, maxY + 1);
  }
}

static void initXrTile(XrTile *tile) {
  initRegion(&tile->dirtyLineArea);
  initRegion(&tile->dirtyRectArea);
  clearXRList(&tile->rects);
  clearXRList(&tile->lines);
}

static void storeLinesInTiles(MaskBuffer *buf) {
  int i,n,m;

  for(i=0; i < buf->lines.used; i++) {
     XSegment *line = ((XSegment*) buf->lines.elements) + i; 

     line->x1 -= buf->region.x;
     line->y1 -= buf->region.y;
     line->x2 -= buf->region.x;
     line->y2 -= buf->region.y;

     int minX, minY, maxX, maxY;
     GET_DIRTY_LINE_REGION(line->x1, line->y1, line->x2, line->y2, minX, minY, maxX, maxY);

     int tileXStartIndex = minX / MASK_TILE_SIZE;
     int tileYStartIndex = minY / MASK_TILE_SIZE;
     int tileXLenght = (maxX / MASK_TILE_SIZE + 1) - tileXStartIndex;
     int tileYLenght = (maxY / MASK_TILE_SIZE + 1) - tileYStartIndex;

      /*TODO: Better coverage analysis should really pay off for lines!*/
      for(n=0; n < tileYLenght; n++) {
        for(m=0; m < tileXLenght; m++) { /*Coverage indicator: Simply do not add to not covered tiles*/
          int tileIndex = (buf->xTileCap * (tileYStartIndex + n)) + tileXStartIndex + m;
          assert(tileIndex >= 0 && tileIndex < (buf->xTileCap*buf->yTileCap));
          XrTile *tile = &buf->tiles[tileIndex];
          XSegment *tileLine = (XSegment*) getNewXR(&tile->lines, sizeof(XSegment));

          int tileStartPosX = (tileXStartIndex + m) * MASK_TILE_SIZE;
          int tileStartPosY = (tileYStartIndex + n) * MASK_TILE_SIZE;

          tileLine->x1 = line->x1 - tileStartPosX; //Translate coordinates relative to the mask-location
          tileLine->y1 = line->y1 - tileStartPosY;
          tileLine->x2 = line->x2 - tileStartPosX;
          tileLine->y2 = line->y2 - tileStartPosY; 

          int transminX = minX - tileStartPosX;
          int transminY = minY - tileStartPosY;
          int transmaxX = maxX - tileStartPosX;
          int transmaxY = maxY - tileStartPosY;

          growDirtyRegionTileLimit(&tile->dirtyLineArea, transminX,  transminY, transmaxX + 1, transmaxY + 1);
       }
     }
  }
}

static void storeRectsInTiles(MaskBuffer *buf) {
  int i,n,m;

  for(i=0; i < buf->rects.used; i++) {
     XRectangle *rect = ((XRectangle*) buf->rects.elements) + i; 

     rect->x -= buf->region.x;
     rect->y -= buf->region.y;

     int tileXStartIndex = rect->x / MASK_TILE_SIZE;
     int tileYStartIndex = rect->y / MASK_TILE_SIZE;
     int tileXLenght = ((rect->x + rect->width) / MASK_TILE_SIZE + 1) - tileXStartIndex;
     int tileYLenght = ((rect->y + rect->height) / MASK_TILE_SIZE + 1) - tileYStartIndex;

      for(n=0; n < tileYLenght; n++) {
        for(m=0; m < tileXLenght; m++) {
          int tileIndex = (buf->xTileCap * (tileYStartIndex + n)) + tileXStartIndex + m;
          assert(tileIndex >= 0 && tileIndex < (buf->xTileCap*buf->yTileCap));

          XrTile *tile = &buf->tiles[tileIndex];
          XRectangle *tileRect = (XRectangle*) getNewXR(&tile->rects, sizeof(XRectangle));

          int tileStartPosX = (tileXStartIndex + m) * MASK_TILE_SIZE;
          int tileStartPosY = (tileYStartIndex + n) * MASK_TILE_SIZE;

          tileRect->x = rect->x - tileStartPosX; //Translate coordinates relative to the mask-location
          tileRect->y = rect->y - tileStartPosY;
          tileRect->width = rect->width;
          tileRect->height = rect->height; 

          limitRetcCoords(tileRect);
          growDirtyRegion(&tile->dirtyRectArea, tileRect->x, tileRect->y, tileRect->x + tileRect->width, tileRect->y + tileRect->height);
       }
     }
  }
}

static void translateRects(XrArrayList *rectList, int x, int y) {
  int i;
  for(i = 0; i < rectList->used; i++) {
      XRectangle *rect = ((XRectangle *) rectList->elements) + i;
      rect->x += x;
      rect->y += y;
   }
}

void fillMask(MaskBuffer* buf, Picture dest) {
  int i, m, n;

  jboolean maskRequired = (xrSrcData.src != xrSrcData.solid) || (xrSrcData.solidColor.alpha != 0xffff) || (xrSrcData.extraAlpha != 1.0f);
  /*Extra Alpha is already stored in solid-color*/
  jdouble maskAlpha = (xrSrcData.src != xrSrcData.texture) ? 1.0f : xrSrcData.extraAlpha;

  unsigned short alphaValueFixed = XDoubleToUShort(maskAlpha);
  XRenderColor maskColor;
  maskColor.alpha = alphaValueFixed;

  /* If we don't have to use a mask and there are no lines,
   * we can draw the rects directly and omit tile partitioning */
  if(maskRequired || buf->lines.used > 0) {
  allocTileBuffer(buf);
  storeRectsInTiles(buf);
  storeLinesInTiles(buf);

  if(buf->lines.used > 0 && buf->validatedGCAlpha != maskAlpha) {
     buf->validatedGCAlpha = maskAlpha;
     XSetForeground(awt_display, buf->drawLineGC, (int) (buf->validatedGCAlpha*255));
  }

  for(i=0; i < buf->yTiles; i++) {
   for(m=0; m < buf->xTiles; m++) {
       XrTile *tile = &buf->tiles[i*buf->xTileCap + m];
       XrArrayList *rectList = &tile->rects;
       XrArrayList *lineList = &tile->lines;

       if(rectList->used != 0 || lineList->used != 0) {
	  MaskRegion tileDirtyArea;
          getTileDirtyArea(&tile->dirtyRectArea, &tile->dirtyLineArea, &tileDirtyArea);

          int tileStartX = m * MASK_TILE_SIZE;
          int tileStartY = i * MASK_TILE_SIZE;
          int x = tileDirtyArea.x + tileStartX + buf->region.x;
          int y = tileDirtyArea.y + tileStartY + buf->region.y;
          int width = tileDirtyArea.x2 - tileDirtyArea.x;
          int height = tileDirtyArea.y2 - tileDirtyArea.y;
          width = MIN(width, MASK_TILE_SIZE);
          height = MIN(height, MASK_TILE_SIZE);

          /*Again, only composite the current tile if required, otherwise just fill rects*/
	  if(maskRequired || lineList->used  > 0) {
            Picture mask = None;

            if(lineList->used > 0) {
              XDrawSegments(awt_display, buf->lineMaskPixmap, buf->drawLineGC, (XSegment *) lineList->elements, lineList->used);
              mask = buf->lineMaskPicture;
            }

            if(rectList->used > 0) {
               if(lineList->used > 0) {
                 XCopyArea(awt_display, buf->lineMaskPixmap, buf->maskPixmap, buf->maskGC, tile->dirtyLineArea.x, tile->dirtyLineArea.y, (tile->dirtyLineArea.x2 - tile->dirtyLineArea.x), (tile->dirtyLineArea.y2 - tile->dirtyLineArea.y), tile->dirtyLineArea.x, tile->dirtyLineArea.y);
                 //XRenderComposite (awt_display, PictOpSrc, buf->lineMaskPicture, None, buf->maskPicture, tile->dirtyLineArea.x, tile->dirtyLineArea.y, 0, 0, tile->dirtyLineArea.x, tile->dirtyLineArea.y, (tile->dirtyLineArea.x2 - tile->dirtyLineArea.x), (tile->dirtyLineArea.y2 - tile->dirtyLineArea.y)); 
               }

               XRenderFillRectangles (awt_display, PictOpSrc, buf->maskPicture, &maskColor, (XRectangle*) rectList->elements, rectList->used);
               mask = buf->maskPicture;
               clearXRList(rectList);
            }

             XRComposite(mask, dest, x, y, tileDirtyArea.x, tileDirtyArea.y, x, y, width, height);

             /* Clear dirty rectangle of the rect-mask */
             XRenderFillRectangle (awt_display, PictOpClear, buf->maskPicture, &color_black, tileDirtyArea.x, tileDirtyArea.y, width, height);

             /* Clear diagonal lines with lines again,
                to avoid the sysmem copy marked "dirty" causing migration for the next lines*/
             if(lineList->used > 0) {
               XDrawSegments(awt_display, buf->lineMaskPixmap, buf->clearLineGC, (XSegment *) lineList->elements, lineList->used);
               clearXRList(lineList);
	     }
          }else
          if(rectList->used > 0) {
             translateRects(rectList, (tileStartX + buf->region.x), (tileStartY + buf->region.y));
             XRenderFillRectangles (awt_display, xrSrcData.compRule, dest, &xrSrcData.solidColor, (XRectangle*) rectList->elements, rectList->used);
             clearXRList(rectList);
          }
        }
      }
     }
  } else {
   XRenderFillRectangles (awt_display, xrSrcData.compRule, dest, &xrSrcData.solidColor, (XRectangle*) buf->rects.elements, buf->rects.used);
 }

  cleanMaskBuffer(buf);
}

#endif
