/*
 * Copyright 2000-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 */

#include <stdlib.h>
#include <jni.h>
#include <jlong.h>
#include "X11SurfaceData.h"
#include "XRSurfaceData.h"
#include "Region.h"
#include "MaskBuffer.h"
#include "Trace.h"

JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRPMBlitLoops_nativeRenderBlit
    (JNIEnv *env, jclass xsd,
     jlong srcData, jlong dstData,
     jint srcx, jint srcy,
     jint dstx, jint dsty,
     jint width, jint height)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRPBMBlitLoops_nativeRenderBlit");

    X11SDOps *srcXsdo, *dstXsdo;
    SurfaceDataBounds span, srcBounds;

    if (width <= 0 || height <= 0) {
        return;
    }

    srcXsdo = (X11SDOps *)jlong_to_ptr(srcData);
    if (srcXsdo == NULL) {
        return;
    }
    dstXsdo = (X11SDOps *)jlong_to_ptr(dstData);
    if (dstXsdo == NULL) {
        return;
    }

    XRenderComposite (awt_display, xrSrcData.compRule, srcXsdo->xrPic, xrSrcData.alphaMask, dstXsdo->xrPic, srcx, srcy, 0, 0, dstx, dsty, width, height); 

    X11SD_DirectRenderNotify(env, dstXsdo);
#endif /* !HEADLESS */
}


JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRPMBlitLoops_nativeTransformedRenderBlit
    (JNIEnv *env, jclass xsd,
     jlong srcData, jlong dstData,
     jint srcx, jint srcy,
     jint dstx, jint dsty,
     jint width, jint height,
     jint m00, jint m01, jint m02, jint m10, jint m11, jint m12, jint maskWidth, jint maskHeight, jint lastWidth, jint lastHeight)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRPBMBlitLoops_nativeTranformedRenderBlit");

    X11SDOps *srcXsdo, *dstXsdo;
    SurfaceDataBounds span, srcBounds;

    if (width <= 0 || height <= 0) {
        return;
    }

    srcXsdo = (X11SDOps *)jlong_to_ptr(srcData);
    if (srcXsdo == NULL) {
        return;
    }
    dstXsdo = (X11SDOps *)jlong_to_ptr(dstData);
    if (dstXsdo == NULL) {
        return;
    }


    Picture mask = None;
    if(maskWidth > -1)
    {
      XRenderColor maskColor;
      maskColor.alpha = XDoubleToUShort(xrSrcData.extraAlpha);

    /*Picture mask;
    switch(maskSize)
    {
      case 1: 
        mask = maskBuffer->img1Mask;
        break;

      case 8:
        mask = maskBuffer->img8Mask;
        break;

      case 64:
        mask = maskBuffer->img64Mask;
        break;
    }*/

   if(maskBuffer->maxWidth < maskWidth || maskBuffer->maxHeight < maskHeight) {
     XRenderPictFormat *fmt = XRenderFindStandardFormat(awt_display, PictStandardA8);
     XRenderPictureAttributes pict_attr;
     Pixmap mp =  XCreatePixmap(awt_display, RootWindow(awt_display, dstXsdo->configData->awt_visInfo.screen), maskWidth, maskHeight, 8);

     //TODO: Alte maske free'n
     maskBuffer->maxMask = XRenderCreatePicture(awt_display, mp, fmt, 0, &pict_attr);
     XRenderFillRectangle (awt_display, PictOpClear, maskBuffer->maxMask, &maskColor, 0, 0, maskWidth, maskHeight);
     maskBuffer->maxWidth = maskWidth;
     maskBuffer->maxHeight = maskHeight;

  //   printf("Reinitialized it!\n"); fflush(stdout);
   }else
   {
    XRenderFillRectangle (awt_display, PictOpClear, maskBuffer->maxMask, &maskColor, 0, 0, lastWidth, lastHeight);
   }

    //printf("Blit: %d, %d, %d, %d\n", maskWidth, maskHeight, lastWidth, lastHeight); fflush(stdout);

    XRenderFillRectangle (awt_display, PictOpSrc, maskBuffer->maxMask, &maskColor, 0, 0, maskWidth, maskHeight);
    
    XTransform tr;
    BUILD_TRANSFORM_MATRIX(tr, m00, m01, m02, m10, m11, m12);
    XRenderSetPictureTransform (awt_display, maskBuffer->maxMask, &tr);

    mask = maskBuffer->maxMask;
}

    XRenderComposite (awt_display, xrSrcData.compRule, srcXsdo->xrPic, mask, dstXsdo->xrPic, srcx, srcy, 0, 0, dstx, dsty, width, height); 
    

    X11SD_DirectRenderNotify(env, dstXsdo);
#endif /* !HEADLESS */
}

#ifndef HEADLESS
/**
 * Prepares a mask-picture according to the mask and the properties currently set, which is used for composition later.
 */
static Picture prepareMaskPM(JNIEnv *env, jbyteArray maskArray, jint maskoff, jint maskscan, jint w, jint h)
{
    int line, pix;
    unsigned char *mask;

    jboolean useEA = (xrSrcData.src == xrSrcData.texture && xrSrcData.extraAlpha != 1.0f);

    if(maskArray != NULL) {

         mask = (unsigned char *) (*env)->GetPrimitiveArrayCritical(env, maskArray, NULL);

         /* Tries to optimize Mask-Upload:  
          * 1. If existing XImage and supplied buffer match, only adjust the data pointer
          * 2. If existing XImage is large enough to hold the data put does not match in scan,
          *    the data is copied in way to fit the XImage.
          * 3. If data is larger than the existing XImage a new temporary XImage is allocated
          * The default XImage is optimized for the AA tiles, which are currently 32x32.
          */
          XImage *defaultImg = maskBuffer->alphaImg;
          XImage *img = maskBuffer->alphaImg;
          jboolean imageFits = defaultImg->width >= w && defaultImg->height >= h;

          if(imageFits && maskoff == defaultImg->xoffset && maskscan == defaultImg->bytes_per_line) {
                defaultImg->data = mask;
          }else {
            if(imageFits) {
               defaultImg->data = maskBuffer->alphaData;

               for(line=0; line < h; line++) {
                   for(pix=0; pix < w; pix++) {
                       img->data[line*img->bytes_per_line + pix] = (unsigned char) (mask[maskscan*line + pix + maskoff]);  
                   }
               }

            }else { 
               /*TODO: handle case where XImage is larger than pixmap*/
               img = XCreateImage(awt_display, &maskBuffer->maskPixmap, 8, ZPixmap, maskoff, mask, maskscan, h, 8, 0);
            }
          }

          /*Upload Image to X-Server*/
          XPutImage(awt_display, maskBuffer->maskPixmap, maskBuffer->maskGC, img, 0, 0, 0, 0, w, h);
          (*env)->ReleasePrimitiveArrayCritical(env, maskArray, mask, JNI_ABORT);

	  if(img != defaultImg) {
             img->data = NULL;
             XDestroyImage(img);
          }

        /* XRender does not expose multitexturing, therefor we have to Composite with the alphaMask to get EA in*/
	if(useEA)  {
                XRenderComposite (awt_display, PictOpSrc, maskBuffer->maskPicture, xrSrcData.alphaMask, maskBuffer->lineMaskPicture, 0, 0, 0, 0, 0, 0, w, h); 
                return maskBuffer->lineMaskPicture;
	}else {
           return maskBuffer->maskPicture;
        }
    }else
    {
      /*If we haven't got a mask (full coverage), we have to return the EA mask, if EA is used*/
      if(useEA) {
        return xrSrcData.alphaMask;
      }
    }

    return None;
}

/* Clean up the area of the masks which have been used as target-pixmap for mask-upload */
static void cleanMask(Picture mask, int w, int h) {
  XRenderFillRectangle (awt_display, PictOpClear, maskBuffer->maskPicture, &color_black, 0, 0, w, h);
  if(mask == maskBuffer->lineMaskPicture) {
     XRenderFillRectangle (awt_display, PictOpClear, maskBuffer->lineMaskPicture, &color_black, 0, 0, w, h);
  }
}
#endif

JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRMaskFill_maskFill
     (JNIEnv *env, jobject joSelf, 
      jlong dstData,
      jint x, jint y, jint w, jint h,
      jint maskoff, jint maskscan, 
      jint masklen, jbyteArray maskArray)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRPBMBlitLoops_maskFill");

    Picture maskPic;
    X11SDOps *dstXsdo;
    dstXsdo = (X11SDOps *)jlong_to_ptr(dstData);
    if (dstXsdo == NULL) {
        return;
    }

    maskPic = prepareMaskPM(env, maskArray, maskoff, maskscan, w, h);
    XRComposite(maskPic, dstXsdo->xrPic, x, y, 0, 0, x, y, w, h);
    cleanMask(maskPic, w, h);

    X11SD_DirectRenderNotify(env, dstXsdo);
#endif
}


JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRMaskBlit_maskBlit
     (JNIEnv *env, jobject joSelf, 
      jlong srcData, jlong dstData,
      jint srcx, jint srcy,
      jint dstx, jint dsty, jint w, jint h,
      jint maskoff, jint maskscan, 
      jint masklen, jbyteArray maskArray)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRPBMBlitLoops_maskBlit");

    Picture maskPic;
    X11SDOps *srcXsdo, *dstXsdo;
    srcXsdo = (X11SDOps *)jlong_to_ptr(srcData);
    dstXsdo = (X11SDOps *)jlong_to_ptr(dstData);
    if (srcXsdo == NULL || dstXsdo == NULL) {
        return;
    }

    maskPic = prepareMaskPM(env, maskArray, maskoff, maskscan, w, h);
    XRenderComposite (awt_display, xrSrcData.compRule, srcXsdo->xrPic, maskPic, dstXsdo->xrPic, srcx, srcy, 0, 0, dstx, dsty, w, h); 
    cleanMask(maskPic, w, h);

    X11SD_DirectRenderNotify(env, dstXsdo);

    #ifdef MITSHM
    if (srcXsdo->isPixmap) {
        X11SD_UnPuntPixmap(srcXsdo);
    }
    #endif /* MITSHM */
#endif
}
