/*
 * Copyright 2001-2003 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 */

#ifndef XRSurfaceData_h_Included
#define XRSurfaceData_h_Included

#define XDoubleToUShort(f)    ((unsigned short) ((f) * 65535))

#define BUILD_TRANSFORM_MATRIX(TRANSFORM, M00, M01, M02, M10, M11, M12)                        \
    {                                                                                          \
      TRANSFORM.matrix[0][0] = M00;                                                            \
      TRANSFORM.matrix[0][1] = M01;                                                            \
      TRANSFORM.matrix[0][2] = M02;                                                            \
      TRANSFORM.matrix[1][0] = M10;                                                            \
      TRANSFORM.matrix[1][1] = M11;                                                            \
      TRANSFORM.matrix[1][2] = M12;                                                            \
      TRANSFORM.matrix[2][0] = 0;                                                              \
      TRANSFORM.matrix[2][1] = 0;                                                              \
      TRANSFORM.matrix[2][2] = 1<<16;                                                          \
    }

/* Holds source-parameters*/
typedef struct {
  Picture src;            /* Currently used source*/
  Picture solid;
  XRenderColor solidColor;
  Picture texture;
  Picture gradient;

  jint compRule;
  jfloat extraAlpha;
  Picture alphaMaskPict;  /*Reference to the 1x1 alpha Mask*/
  Picture alphaMask;      /*Alpha-Mask which should be used for composition, either alphaMaskPict or None*/ 

  jboolean gradCacheEnabled;
  Pixmap gradPixmap;
  Picture gradPict;
} SrcSurfaceData;

extern SrcSurfaceData xrSrcData;
extern MaskBuffer *maskBuffer;
extern XRenderColor color_black;

/* Composite using gradient-caching if enabled */ 
void XRComposite(Picture mask, Picture dest, int srcX, int srcY, int maskX, int maskY, int dstX, int dstY, unsigned int width, unsigned int height);

/* Initializes the global source structure*/
static void initXRSrcData();

/*Decodes non-premultiplied pixel, multiplies it and stores the result in color*/
static void decodeRenderColor(jint pixel, XRenderColor *color);

/*Decodes a pixel and stores the result in color*/
static void decodeRenderColorPre(jint pixel, XRenderColor *color);

#endif //XRSurfaceData_h_Included
